set STATES_MAX 10


proc error_detected {s} {
    puts "ERROR $s"
    exit 1
}



puts "
#ifndef __METRICS_DATA_H
#define __METRICS_DATA_H
"


set fd [open "metrics_definitions.data"]
set data [read $fd]

set tr_id 0

set out_data {}
set names {}
foreach metric $data {
    set metric_name [lindex $metric 0]
    set description [lindex $metric 1]
    set kind        [lindex $metric 2]
    set states_list [lindex $metric 3]

    set out_metric [list $metric_name $description $kind $states_list]

    set transition_list [lindex $metric 4]

    lappend names $metric_name

    set states_names {}
    foreach item $states_list {
	lappend states_names    [lindex $item 0]
    }
    
    set NUM_STATES [llength $states_list]

    for {set i 0} {$i < $NUM_STATES} {incr i} {
	set table($i) {}
    }


    set transition_ptr_list {}

    foreach t $transition_list {
	
	set from [lsearch $states_names [lindex $t 0]]
	
	if { $from < 0 } {
	    error_detected "in $t. From state not defined"
	}
	
	set to   [lsearch $states_names [lindex $t 1]]

	if { $to < 0 } {
	    error_detected "in $t. To state not defined"
	}
	
	set label      [lindex $t 2]

	lappend table($from) [list $to $label] 
    }


    puts "//  Metric $metric_name"
    puts ""
    
    for {set i 0} {$i < $NUM_STATES} {incr i} {

	set transition_number 0

	foreach t $table($i) { 

	    if {$t != "NO_TRANSITION"} {
		
		set to    [lindex $t 0]
		set label [lindex $t 1]
		
		set kind [lindex $label 0]
		
			
		switch  $kind {
		    LAMBDA {
			set event "M_ANY_EVENT"
			set thread "ANY_THREAD"
			set object "ANY_OBJECT"
		    }
		    
		    NORMAL {
			set event "[lindex $label 1]"
			
			set thread [lindex $label 2]
			
			switch -regexp $thread {
			    {SELF_THREAD|ANY_THREAD|OTHER_THREAD} {
			    }
			    default {
				error_detected "in $t ($thread)"
			    }
			}
			
			set object [lindex $label 3]
			switch -regexp $object {
			    {SELF_OBJECT|ANY_OBJECT|OTHER_OBJECT} {
			    }
			    default {
				error_detected "in $t ($object)"
			    }
			}
			
			
		    }
		    default {
			error_detected "in $t ($kind)"
		    }
		}
		
		if {$transition_number == 0} {
		    set next NULL
		} else {
		    set next "&tr_$tr_id"
		}
		incr tr_id
		puts "mtri_transition_descriptor_t tr_$tr_id = \n      \{$to, $kind, $event, $thread, $object, $next\};"
		
		incr transition_number
		
	    }
	}
	
	if {$transition_number > 0} {
	    lappend transition_ptr_list $tr_id
	} else {
	    error_detected "STATE [lindex $states_names $i] without out transition"
	}
    }

    lappend out_metric $transition_ptr_list

    lappend out_data $out_metric
    puts ""
}


set NUM_METRICS [llength $names]

set i 0
foreach metric_name $names {
    puts [format "#define %-32s %d" $metric_name $i]
    incr i
}

puts ""

puts "mtri_metric_descriptor_t mtri_metric \[\] ="
puts "\{"

set n 0
foreach this_metric $out_data {


    puts "\{"
     
    set name [lindex $this_metric 0]
    set description [lindex $this_metric 1]
    set kind [lindex $this_metric 2]
    set states_desc [lindex $this_metric 3]
    set transition_ptr_list [lindex $this_metric 4]

    puts "  \"$name\","    
    puts "  \"$description\","
    
    puts "\{"
    
    puts "$name,"    
    puts "$kind,"

    set states {}
    foreach s $states_desc {
	lappend states [lindex $s 0]
    }
    
    set NUM_STATES [llength $states_desc]

    puts "$NUM_STATES,"

    puts -nonewline "\{"
    set i 0
    while {$i < $NUM_STATES} {
	set sd [lindex $states_desc $i]
	set status_class [lindex $sd 1]
	puts -nonewline "\{$status_class\}"
	incr i
	if {$i != $NUM_STATES} {
	    puts -nonewline ", "
	}
    }
    puts "\},"
    


    puts -nonewline "\{"
    set i 0
    while {$i < $NUM_STATES} {
	set td [lindex $transition_ptr_list $i]
	set ptr [lindex $td 0]
	puts -nonewline "&tr_$ptr"
	incr i
	if {$i != $NUM_STATES} {
	    puts -nonewline ", "
	}
    }
    puts "\}"
    puts "\}"
    
    incr n
    if {$n < $NUM_METRICS} {
	puts "\},"
    } else {
	puts "\}"
    }


}

puts "\};"



puts "
#define NUM_SYSTEM_METRICS $NUM_METRICS

#endif // __METRICS_DATA_H

"