#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <expat.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#define TRUE  1
#define FALSE 0

#define NO_METRIC  -1
#define NO_THREAD  -1
#define NO_OBJECT  -1

#define NS_IN_SEC 1000000000LL
#define INFINITY 0x7fffFfffFfffFfffLL

#define CDATA_LEN_MAX   256
#define ELEMENT_LEN_MAX  64
#define INDEX_MAX        32
#define METRICS_MAX     256
#define DIST_MAX         21
#define PATH_MAX        512

typedef long long nstime;



typedef struct {
  int             metric_id;
  int             thread_id;
  int             object_id;
  nstime          duration;
  nstime          begin;
  nstime          end;
  int             events_count;
  int             id;
  void *          next;
} measurement_t;


typedef struct {
  int             n;
  nstime          sum_x;
  double          sum_x2;
  nstime          min;
  nstime          max;
  int             v [DIST_MAX];
  nstime          ampli;
  nstime          first;
  nstime          last;

} dist_t;


typedef struct {
  int             index1;
  int             index2;
  measurement_t * first;
  measurement_t * last;

  dist_t          dist;

} measurement_list_t;



typedef struct {
  int                used; 
  char               metric_name [CDATA_LEN_MAX];
  char               metric_description [CDATA_LEN_MAX];
  int                with_metric_summary;
  int                with_specific_distributions;
  int                show_thread;
  int                show_object;
  dist_t             metric_dist;
  dist_t             one_level_dist [INDEX_MAX];
  measurement_list_t measurement_list [INDEX_MAX][INDEX_MAX];

} metric_t;


struct {
  int  metric_id;
  char metric_name [CDATA_LEN_MAX];
  char metric_description [CDATA_LEN_MAX];
} current_metric;

struct {
  char created [CDATA_LEN_MAX];
  char host    [CDATA_LEN_MAX];
  char user    [CDATA_LEN_MAX];


} header;


metric_t metric [METRICS_MAX];

char current_cdata [CDATA_LEN_MAX];
int  current_cdata_len = 0;

measurement_t current_measurement;

int in_MEASUREMENTS = FALSE;

int must_make_measurements  = FALSE;

char dir [PATH_MAX] = ".";
char dir_figs [PATH_MAX];
char gnuplot_command [PATH_MAX];
char progfilename [PATH_MAX];
char inputfilename [PATH_MAX] = "";
char reportfilename [PATH_MAX];


FILE * report;

//////////////////////////////////////////////////////////////
//  MISC FUNCTIONS
//////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////
//  usage
//////////////////////////////////////////////////////////////


static void
usage (char * s)
{
  fprintf (stderr, "usage: %s [-m] [-i input] [-d directory] [--help]\n", s);
  fprintf (stderr, "\t -m           : make measurements chapter\n");
  fprintf (stderr, "\t -i input     : input file (default standard input)\n");
  fprintf (stderr, "\t -d directory : output directory (default current directory)\n");
  fprintf (stderr, "\t --help       : prints this message and exit\n");
  exit (1);
}


//////////////////////////////////////////////////////////////
//  standard_deviation
//////////////////////////////////////////////////////////////


static double
standard_deviation (int    n,
		    nstime sum_x,
		    double sum_x2)
{

  double var;
  double s_sx;


  s_sx  = (double) sum_x / 1000000000.0;

  var = (sum_x2 - (s_sx * s_sx) / n) / n;
  if (var == 0.0) {
    return 0.0;
  } else {
    return (sqrt (var));
  }
}


//////////////////////////////////////////////////////////////
//  string_to_ns
//////////////////////////////////////////////////////////////

nstime
string_to_ns (char *from) {

  int sec, nsec;
  sscanf (from, "%d.%d", &sec, &nsec);

  return ((nstime)sec * (nstime)1000000000 + (nstime)nsec);

}


//////////////////////////////////////////////////////////////
//  initialize_dist
//////////////////////////////////////////////////////////////

static void
initialize_dist (dist_t * dist) {

  int k;

  dist->n      = 0;
  dist->sum_x  = 0;
  dist->sum_x2 = 0;
  dist->min    = INFINITY;
  dist->max    = 0;
  for (k = 0; k < INDEX_MAX; k++) {
   dist->v [k] = 0;
  }
}

//////////////////////////////////////////////////////////////
//  initialize
//////////////////////////////////////////////////////////////

static void
initialize (void) {

  int i;
  int j;
  int k;

  for (i = 0; i < METRICS_MAX; i++) {
    metric [i].with_metric_summary = TRUE;
    metric [i].with_specific_distributions = TRUE;
    metric [i].show_thread = TRUE;
    metric [i].show_object = FALSE;
    initialize_dist (&metric[i].metric_dist);

    for (j = 0; j < INDEX_MAX; j++) {
      initialize_dist (&metric[i].one_level_dist [j]);
      for (k = 0; k < INDEX_MAX; k++) {
	initialize_dist (&(metric[i].measurement_list [j][k].dist));
      }
    }    
  }

}


static void
calculate_distributions (void) {

  int i;
  int j;
  int k;
  int l;
  metric_t           * m;
  measurement_list_t * list;
  measurement_t      * ptr;
  nstime               ampli;


  
  
  for (i = 0; i < METRICS_MAX; i++) {

    if (metric [i].used == TRUE) {

      m = & (metric [i]);

      ampli = (m->metric_dist.max - m->metric_dist.min) / (DIST_MAX - 1);
      ampli++;

	
      m->metric_dist.ampli = ampli;

      m->metric_dist.first = m->metric_dist.min - ampli / 2;
      m->metric_dist.last  = m->metric_dist.first + ampli * DIST_MAX;

      for (j = 0; j < INDEX_MAX; j++) {
	for (k = 0; k < INDEX_MAX; k++) {
	  list = & (m->measurement_list [j][k]);
    
	  ptr = list->first;
    
	  while (ptr != NULL) {
	    l = (ptr->duration - m->metric_dist.first) / ampli;

	    /* 	  if (k >= 21) { */
	    /* 	    printf ("ALERTA"); */
	    /* 	    exit (1); */
	    /* 	  } */

	    (m->metric_dist.v [l])++;
	    ptr = ptr->next;
	  }
	}
      }
    }
 
  }
  
  
  for (i = 0; i < METRICS_MAX; i++) {

    if (metric [i].used == TRUE) {

      m = & (metric [i]);

      for (j = 0; j < INDEX_MAX; j++) {
	for (k = 0; k < INDEX_MAX; k++) {
	  list = & (m->measurement_list [j][k]);

	  ampli = (list->dist.max - list->dist.min) / (DIST_MAX - 1);
	  ampli++;

	  //if (ampli % 2 == 0) {
	  //  ampli++;
	  //}
	
	  list->dist.ampli = ampli;

	  list->dist.first = list->dist.min - ampli / 2;
	  list->dist.last  = list->dist.first + ampli * DIST_MAX;

	  ptr = list->first;
    
	  while (ptr != NULL) {
	    l = (ptr->duration - list->dist.first) / ampli;

	    if (l >= 21) {
	      printf ("ALERTA");
	      exit (1);
	    }

	    (list->dist.v [l])++;
	    ptr = ptr->next;
	  }
	}
      }
    }    
  }
  
}


//////////////////////////////////////////////////////////////
//  DOCBOOK
//////////////////////////////////////////////////////////////


static void
make_distribution (dist_t *dist,
		   int metric_id,
		   int object_id,
		   int thread_id) {


  FILE * prog;
  int k;

  nstime mid_point;
  nstime first;
  nstime last;
  int    yrange;
  int    ymax   = 0;

  char copy [PATH_MAX];
  char temp [PATH_MAX];
  char filepng [PATH_MAX];
  char fileps [PATH_MAX];
  char reffilepng [PATH_MAX];
  char reffileps [PATH_MAX];

  first = dist->first - dist->ampli;
  last  = dist->last + dist->ampli;

  for (k = 0; k < DIST_MAX; k++) {
    if (dist->v [k] > ymax) {
      ymax = dist->v [k];
    }
  }

  yrange = (ymax + 1) * 1.1;

  prog = fopen (progfilename, "w+");

  sprintf (filepng, "%s/fig-%d-%d-%d.png", dir_figs, metric_id, object_id, thread_id);
  sprintf (fileps, "%s/fig-%d-%d-%d.ps", dir_figs, metric_id, object_id, thread_id);
  sprintf (reffilepng, "figs/fig-%d-%d-%d.png", metric_id, object_id, thread_id);
  sprintf (reffileps, "figs/fig-%d-%d-%d.ps", metric_id, object_id, thread_id);

  fprintf (prog, "set format x '%%.9f'\n");
  fprintf (prog, "set rmargin 8\n");
  fprintf (prog, "set lmargin 8\n");
  fprintf (prog, "set output '%s'\n", filepng);
  // fprintf (prog, "set terminal png xfefec8\n");
  fprintf (prog, "set terminal png xffffff x000000  x000000 x0000ff   \n");
  fprintf (prog, "set nokey\n");
  fprintf (prog, "set xtics ('%lld.%09lld' %lld.%09lld, '%lld.%09lld' %lld.%09lld)\n", 
	   dist->min / NS_IN_SEC,
	   dist->min % NS_IN_SEC,
	   dist->min / NS_IN_SEC,
	   dist->min % NS_IN_SEC,
	   dist->max / NS_IN_SEC,
	   dist->max % NS_IN_SEC,
	   dist->max / NS_IN_SEC,
	   dist->max % NS_IN_SEC);
  fprintf (prog, "set ytics (%d, %d)\n", ymax, ymax / 2);
  fprintf (prog, "set size nosquare 0.6,0.3\n");

  if (first >= 0) {
    fprintf (prog, "plot [%lld.%09lld:%lld.%09lld] [%d:%d] '-' with boxes\n",
	     first / NS_IN_SEC,
	     first % NS_IN_SEC,
	     last / NS_IN_SEC,
	     last % NS_IN_SEC,
	     0,
	     yrange);
  } else {
    fprintf (prog, "plot [-%lld.%09lld:%lld.%09lld] [%d:%d] '-' with boxes\n",
	     -first / NS_IN_SEC,
	     -first % NS_IN_SEC,
	     last / NS_IN_SEC,
	     last % NS_IN_SEC,
	     0,
	     yrange);
  }

  
  for (k = 0; k < DIST_MAX; k++) {
    mid_point = dist->first + dist->ampli * (nstime) k + dist->ampli / 2;

    fprintf (prog, "%lld.%09lld %d\n",
	     mid_point / NS_IN_SEC,
	     mid_point % NS_IN_SEC,
	     dist->v [k]);
  }

  fprintf (prog, "e\n");
  fprintf (prog, "set output '%s'\n", fileps);
  fprintf (prog, "set terminal postscript eps\n");
  fprintf (prog, "set size nosquare 0.4,0.2\n");
  fprintf (prog, "replot\n");
  
  fclose (prog);

  system (gnuplot_command);

  
  fprintf (report, "<mediaobject>\n");
  fprintf (report, "<imageobject>\n");
  fprintf (report, "<imagedata fileref=\"%s\"></imagedata>\n", reffilepng);
  fprintf (report, "</imageobject>\n");
  fprintf (report, "<imageobject>\n");
  fprintf (report, "<imagedata fileref=\"%s\"></imagedata>\n", reffileps);
  fprintf (report, "</imageobject>\n");
   fprintf (report, "</mediaobject>\n");

  // system (pnm_command);

  sprintf (copy,  "%s/copia.%d-%d-%d", dir_figs, metric_id, object_id, thread_id);
  strcpy (temp, "cp ");
  strcat (temp, progfilename);
  strcat (temp, " ");
  strcat (temp, copy);

  system (temp);






}



static void
make_measurements (int metric_id)

{

  metric_t           * m;
  measurement_list_t * list;
  int                  i;
  int                  j;
  measurement_t      * ptr;
  int                  num_measurements = 0;

  m = & (metric [metric_id]);

  fprintf (report, "\n");
  fprintf (report, "<section id='measurements-%d'>\n", metric_id);
  fprintf (report, "<title id='measurements-%d.title'>Measurements for Metric %s</title>\n", metric_id, m->metric_name);

  fprintf (report, "<para><emphasis>Metric Description</emphasis>: %s</para>\n", m->metric_description);



  fprintf (report, "<table frame='all' tocentry='0'>\n");
  fprintf (report, "<title>Measurements Clasified by Thread Identifier</title>\n");
  
  fprintf (report, "  <tgroup cols='6'>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
  fprintf (report, "    <thead>\n");
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry>Thread Id</entry>\n"); 
  fprintf (report, "        <entry>Object Id</entry>\n"); 
  fprintf (report, "        <entry>Duration</entry>\n"); 
  fprintf (report, "        <entry>Begin</entry>\n"); 
  fprintf (report, "        <entry>End</entry>\n"); 
  fprintf (report, "        <entry>Events Detected</entry>\n"); 
  fprintf (report, "        <entry>Id</entry>\n"); 
  fprintf (report, "        </row>\n"); 
  fprintf (report, "    </thead>\n");
  
  fprintf (report, "    <tbody>\n");

  for (i = 0; i < INDEX_MAX; i++) {
    for (j = 0; j < INDEX_MAX; j++) {
    
      list = & (m->measurement_list [i][j]);

      ptr = list->first;
    
      while (ptr != NULL) {
	fprintf (report, "      <row>\n");
	if (ptr->thread_id == NO_THREAD) {
	  fprintf (report, "        <entry>--</entry>\n");
	} else {
	  fprintf (report, "        <entry>%d</entry>\n"  , ptr->thread_id);
	}
	if (ptr->object_id == NO_OBJECT) {
	  fprintf (report, "        <entry>--</entry>\n");
	} else {
	  fprintf (report, "        <entry>%d</entry>\n"  , ptr->object_id);
	}
	fprintf (report, "        <entry>%lld.%09lld</entry>\n", ptr->duration / NS_IN_SEC, ptr->duration % NS_IN_SEC);
	fprintf (report, "        <entry>%lld.%09lld</entry>\n", ptr->begin / NS_IN_SEC, ptr->begin % NS_IN_SEC);
	fprintf (report, "        <entry>%lld.%09lld</entry>\n", ptr->end / NS_IN_SEC, ptr->end % NS_IN_SEC);
	fprintf (report, "        <entry>%d</entry>\n", ptr->events_count);
	fprintf (report, "        <entry>%d</entry>\n", ptr->id);
	fprintf (report, "      </row>\n"); 
	ptr = ptr->next;

	num_measurements++;
      }
    }
  }

  if (num_measurements == 0) {
    fprintf (report, "      <row>\n");
    fprintf (report, "        <entry>--</entry>\n");
    fprintf (report, "        <entry>--</entry>\n");
    fprintf (report, "        <entry>--</entry>\n");
    fprintf (report, "        <entry>--</entry>\n");
    fprintf (report, "        <entry>--</entry>\n");
    fprintf (report, "        <entry>--</entry>\n");
    fprintf (report, "        <entry>--</entry>\n");
    fprintf (report, "      </row>\n"); 
  }
  
  fprintf (report, "    </tbody>\n");
  fprintf (report, "  </tgroup>\n");
  fprintf (report, "</table>\n");


  fprintf (report, "<para>The distributions for this metric are available in section <link linkend='metric-%d' endterm=\"metric-%d.title\"></link></para>", metric_id, metric_id);
  fprintf (report, "</section>\n");
}

static void
make_chapter_measurements (void) 
{

  int i;

  fprintf (report, "\n");
  fprintf (report, "<chapter>\n");
  fprintf (report, "<title>Measurements Retrieved</title>\n");

  for (i = 0; i < METRICS_MAX; i++) {
    if (metric [i].used == TRUE) {
      make_measurements (i);
    }
  }    
  fprintf (report, "</chapter>\n");
  fprintf (report, "\n");
}




static void
make_metric_results (int metric_id)

{

  metric_t           * m;
  measurement_list_t * list;
  int                  i;
  int                  j;
  // int                  k;
  nstime               med;

  m = & (metric [metric_id]);

  fprintf (report, "\n");
  fprintf (report, "<section id='metric-%d'>\n", metric_id);
  fprintf (report, "<title id='metric-%d.title'>Distributions for Metric %s</title>\n", metric_id, m->metric_name);
  
  fprintf (report, "<para><emphasis>Metric Description</emphasis>: %s</para>\n", m->metric_description);

  if (m->metric_dist.n == 0) {
    fprintf (report, "<para>No measurements found</para>\n");
    fprintf (report, "</section>\n");
    return;
  }


  if (m->with_metric_summary == TRUE) {

    fprintf (report, "<table frame='all' tocentry='0'>\n");
    fprintf (report, "<title>Overall Distribution</title>\n");
  
    fprintf (report, "  <tgroup cols='6'>\n");
    fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
    fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
    fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
    fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
    fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
    fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
    fprintf (report, "    <thead>\n");
    fprintf (report, "      <row>\n"); 
    fprintf (report, "        <entry>N</entry>\n"); 
    fprintf (report, "        <entry>Min</entry>\n"); 
    fprintf (report, "        <entry>Max</entry>\n"); 
    fprintf (report, "        <entry>Max - Min</entry>\n"); 
    fprintf (report, "        <entry>Mean</entry>\n"); 
    fprintf (report, "        <entry>Std. Dev.</entry>\n"); 
    fprintf (report, "        </row>\n"); 
    fprintf (report, "    </thead>\n");
  
    fprintf (report, "    <tbody>\n");

    fprintf (report, "    <row>\n"); 
    fprintf (report, "      <entry>%d</entry>\n", m->metric_dist.n);
    fprintf (report, "      <entry>%lld.%09lld</entry>\n", m->metric_dist.min / NS_IN_SEC, m->metric_dist.min % NS_IN_SEC);
    fprintf (report, "      <entry>%lld.%09lld</entry>\n", m->metric_dist.max / NS_IN_SEC, m->metric_dist.max % NS_IN_SEC);
    fprintf (report, "      <entry>%lld.%09lld</entry>\n", (m->metric_dist.max - m->metric_dist.min) / NS_IN_SEC, (m->metric_dist.max  - m->metric_dist.min) % NS_IN_SEC);

    med = m->metric_dist.sum_x / m->metric_dist.n;
    fprintf (report, "      <entry>%lld.%09lld</entry>\n", med / NS_IN_SEC, med % NS_IN_SEC);
    //fprintf (report, "      <entry> - </entry>\n");
    fprintf (report, "      <entry>%.9f</entry>\n", standard_deviation (m->metric_dist.n, m->metric_dist.sum_x, m->metric_dist.sum_x2));
    fprintf (report, "    </row>\n"); 


    fprintf (report, "    </tbody>\n");
    fprintf (report, "  </tgroup>\n");
    fprintf (report, "</table>\n");

    make_distribution (&(m->metric_dist), metric_id, NO_OBJECT, NO_THREAD);
  } else {
    fprintf (report, "<para> Overall distribution is not applicable to this metric</para>");
  }

  if (m->with_specific_distributions == TRUE) {


    if (m->show_object) {
      fprintf (report, "<para> Detailed distributions, clasified by object, follow.</para>");
    } else {
      fprintf (report, "<para> Detailed distributions, clasified by thread, follow.</para>");
    }
    for (i = 0; i < INDEX_MAX; i++) {
      for (j = 0; j < INDEX_MAX; j++) {
	list = & (m->measurement_list [i][j]);
	if (list->first != NULL) {
	  fprintf (report, "<table frame='all' tocentry='0'>\n");

	  fprintf (report, "<title>");
	  if (m->show_object) {
	    fprintf (report, "Object %d, ", i);
	  }
	  if (m->show_thread) {
	    fprintf (report, "Thread %d", j);
	  }
	  fprintf (report, " Distribution</title>\n");
	  
  
	  fprintf (report, "  <tgroup cols='6'>\n");
	  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
	  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
	  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
	  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
	  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
	  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
	  fprintf (report, "    <thead>\n");
	  fprintf (report, "      <row>\n"); 
	  fprintf (report, "        <entry>N</entry>\n"); 
	  fprintf (report, "        <entry>Min</entry>\n"); 
	  fprintf (report, "        <entry>Max</entry>\n"); 
	  fprintf (report, "        <entry>Max - Min</entry>\n"); 
	  fprintf (report, "        <entry>Mean</entry>\n"); 
	  fprintf (report, "        <entry>Std. Dev.</entry>\n"); 
	  fprintf (report, "        </row>\n"); 
	  fprintf (report, "    </thead>\n");
  
	  fprintf (report, "    <tbody>\n");

	  fprintf (report, "    <row>\n"); 
	  fprintf (report, "      <entry>%d</entry>\n", list->dist.n); 
	  fprintf (report, "      <entry>%lld.%09lld</entry>\n", list->dist.min / NS_IN_SEC, list->dist.min % NS_IN_SEC);
	  fprintf (report, "      <entry>%lld.%09lld</entry>\n", list->dist.max / NS_IN_SEC, list->dist.max % NS_IN_SEC);
	  fprintf (report, "      <entry>%lld.%09lld</entry>\n", (list->dist.max - list->dist.min) / NS_IN_SEC, (list->dist.max  - list->dist.min) % NS_IN_SEC);

	  med = list->dist.sum_x / list->dist.n;
  
	  fprintf (report, "      <entry>%lld.%09lld</entry>\n", med / NS_IN_SEC, med % NS_IN_SEC);
	  // fprintf (report, "      <entry> - </entry>\n");
	  fprintf (report, "      <entry>%.9f</entry>\n", standard_deviation (list->dist.n, list->dist.sum_x, list->dist.sum_x2));
	  fprintf (report, "    </row>\n"); 

	  fprintf (report, "    </tbody>\n");
	  fprintf (report, "  </tgroup>\n");
	  fprintf (report, "</table>\n");

	  /*       fprintf (report, "<para>\n"); */
	  /*       for (k = 0; k < DIST_MAX; k++) { */
	  /* 	fprintf (report, "%d,", list->dist.v [k]); */
	  /*       } */
	  /*       fprintf (report, "</para>\n"); */

	  make_distribution (&(list->dist), metric_id, i, j);
	}
      }
    }
  } else {
    fprintf (report, "<para> Detailed distributions are not applicable to this metric</.para>");
  }

  if (must_make_measurements == TRUE) {
    fprintf (report, "<para>Measurements for this metric are available in section <link linkend='measurements-%d' endterm=\"measurements-%d.title\"></link></para>", metric_id, metric_id);
  }

  fprintf (report, "</section>\n");

}


static void
make_section_utilization (void) 
{
  nstime interval;
  nstime begin = INFINITY;
  nstime end   = 0LL;
  double total_utilization = 0.0;
  double utilization;
  double idle_utilization = 0.0;

  metric_t           * m;
  measurement_list_t * list;
  measurement_t      * ptr;

  int    i;
  int    j;
  int running_global_id = -1;



  for (i = 0; i < METRICS_MAX; i++) {
    
    if (metric [i].used == TRUE) {
      if (strcmp (metric [i].metric_name, "RUNNING_GLOBAL") == 0) {
	running_global_id = i;
      }
    }
  } // for

  if (running_global_id == -1) {
    fprintf (report, "<section>\n");
    fprintf (report, "<title>Processor Utilization </title>\n");
    fprintf (report, "<para>Proccesor Utilization depends on metric RUNNING_GLOBAL which is not available</para>\n");
    fprintf (report, "</section>\n");    
    return;
  }

  
  m = & (metric [running_global_id]);

  for (j = 0; j < INDEX_MAX; j++) {

    list = & (m->measurement_list [0][j]);
	 
    ptr = list->first;
    
    while (ptr != NULL) {

      if (ptr->begin < begin) {
	begin = ptr->begin;
      }
      if (ptr->end > end) {
	end = ptr->end;
      }
      ptr = ptr->next;
    }

  }

  interval = end - begin;

  
  fprintf (report, "<section>\n");
  fprintf (report, "<title>Processor Utilization </title>\n");


  fprintf (report, "<table frame='all' tocentry='0'>\n");
  fprintf (report, "<title></title>\n");
  
  fprintf (report, "  <tgroup cols='2'>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right' colname='c1'></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right' colname='c2'></colspec>\n");
  fprintf (report, "    <thead>\n");
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry>Thread Id</entry>\n"); 
  fprintf (report, "        <entry>Utilization</entry>\n"); 
  fprintf (report, "      </row>\n"); 
  fprintf (report, "    </thead>\n");
  
  fprintf (report, "    <tbody>\n");


  for (j = 0; j < INDEX_MAX; j++) {

    list = & (m->measurement_list [0][j]);
	
    ptr = list->first;
    
    while (ptr != NULL) {

      if (ptr->thread_id == 0) {

	idle_utilization = (double) ptr->duration / (double) interval;

      } else {
      
	utilization = (double) ptr->duration / (double) interval;
	total_utilization = total_utilization + utilization;

	fprintf (report, "    <row>\n"); 
	fprintf (report, "      <entry>%d</entry>\n", ptr->thread_id);
	fprintf (report, "      <entry>%f</entry>\n", utilization * 100);
	fprintf (report, "    </row>\n"); 
      }
      
      ptr = ptr->next;
    }
    
  }
  

  fprintf (report, "    </tbody>\n");
  fprintf (report, "  </tgroup>\n");

  fprintf (report, "  <tgroup cols='2'>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right' colname='c1'></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right' colname='c2'></colspec>\n");
  fprintf (report, "    <spanspec spanname='span' namest='c1' nameend='c2' align='center'></spanspec>\n");
  fprintf (report, "    <thead>\n");
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry spanname='span'>Totals</entry>\n"); 
  fprintf (report, "      </row>\n"); 
  fprintf (report, "    </thead>\n");
  fprintf (report, "    <tbody>\n");
  fprintf (report, "    <row>\n"); 
  fprintf (report, "      <entry>System</entry>\n");
  fprintf (report, "      <entry>%f</entry>\n", total_utilization * 100);
  fprintf (report, "    </row>\n"); 
  fprintf (report, "    <row>\n"); 
  fprintf (report, "      <entry>Idle (Linux)</entry>\n");
  fprintf (report, "      <entry>%f</entry>\n", idle_utilization * 100);
  fprintf (report, "    </row>\n"); 
  fprintf (report, "    </tbody>\n");
  fprintf (report, "  </tgroup>\n");
 
  fprintf (report, "</table>\n");
  fprintf (report, "</section>\n");


}
    




static void
make_chapter_additional (void) 
{

  fprintf (report, "\n");
  fprintf (report, "<chapter>\n");
  fprintf (report, "<title>Aditional Information</title>\n");

  make_section_utilization ();

  fprintf (report, "</chapter>\n");
  fprintf (report, "\n");
}


static void
make_chapter_metrics (void) 
{

  int i;

  fprintf (report, "\n");
  fprintf (report, "<chapter>\n");
  fprintf (report, "<title>Metrics Distributions</title>\n");

  for (i = 0; i < METRICS_MAX; i++) {
    if (metric [i].used == TRUE) {
      make_metric_results (i);
    }
  }    
  fprintf (report, "</chapter>\n");
  fprintf (report, "\n");
}



static void
make_chapter_summary (void) 
{

  int i;

  fprintf (report, "\n");
  fprintf (report, "<chapter>\n");
  fprintf (report, "<title>Introduction</title>\n");

  fprintf (report, "<para> This is a report generated by the \"todocbook\" command </para>\n");
  
  fprintf (report, "<informaltable frame='all'>\n");
  fprintf (report, "  <tgroup cols='2'>\n");
  fprintf (report, "    <colspec colwidth=\"1*\"></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");

  fprintf (report, "    <tbody>\n");
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry>Date of Creation</entry>\n"); 
  fprintf (report, "        <entry>%s</entry>\n", header.created); 
  fprintf (report, "      </row>\n"); 
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry>User</entry>\n"); 
  fprintf (report, "        <entry>%s</entry>\n", header.user); 
  fprintf (report, "      </row>\n"); 
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry>Host</entry>\n"); 
  fprintf (report, "        <entry>%s</entry>\n", header.host); 
  fprintf (report, "      </row>\n"); 
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry>Time Units</entry>\n"); 
  fprintf (report, "        <entry>Seconds</entry>\n"); 
  fprintf (report, "      </row>\n"); 
  fprintf (report, "    </tbody>\n");
  fprintf (report, "  </tgroup>\n");
  fprintf (report, "</informaltable>\n");

  
  fprintf (report, "<table frame='all' tocentry='0'>\n");
  fprintf (report, "<title>Metrics requested</title>\n");
  
  fprintf (report, "  <tgroup cols='2'>\n");
  fprintf (report, "    <colspec colwidth=\"1*\"></colspec>\n");
  fprintf (report, "    <colspec colwidth=\"1*\" align='right'></colspec>\n");
  fprintf (report, "    <thead>\n");
  fprintf (report, "      <row>\n"); 
  fprintf (report, "        <entry>Metric Name</entry>\n"); 
  fprintf (report, "        <entry>Measurements Retrieved</entry>\n"); 
  fprintf (report, "        </row>\n"); 
  fprintf (report, "    </thead>\n");
  
  fprintf (report, "    <tbody>\n");

  for (i = 0; i < METRICS_MAX; i++) {
    if (metric [i].used == TRUE) {
      fprintf (report, "    <row>\n"); 
      fprintf (report, "      <entry><link linkend='metric-%d'>%s</link></entry>\n", i, metric [i].metric_name); 
      fprintf (report, "      <entry>%d</entry>\n", metric [i].metric_dist.n); 
      fprintf (report, "    </row>\n"); 
    }

  }

  fprintf (report, "    </tbody>\n");
  fprintf (report, "  </tgroup>\n");
  fprintf (report, "</table>\n");


  fprintf (report, "</chapter>\n");
  fprintf (report, "\n");
  
}

static void
make_docbook (void) 
{
  


  fprintf (report, "<!DOCTYPE book PUBLIC \"-//OASIS//DTD DocBook V3.1//EN\" \"http://www.oasis-open.org/docbook/xml/4.2/docbookx.dtd\">\n");
  fprintf (report, "<book lang=\"en\">\n");
  fprintf (report, "<title>Metric Report</title>\n");

  make_chapter_summary ();
  make_chapter_metrics ();
  make_chapter_additional ();
  if (must_make_measurements == TRUE) {
    make_chapter_measurements ();
  }

  fprintf (report, "</book>\n");
}



//////////////////////////////////////////////////////////////
//  ELEMENT START
//////////////////////////////////////////////////////////////

static void
start_METRICSREQUESTED (void) {

  in_MEASUREMENTS = FALSE;

}




static void
start_MEASUREMENTS (void) {

  in_MEASUREMENTS = TRUE;

}







//////////////////////////////////////////////////////////////
//  ELEMENT FINISHED
//////////////////////////////////////////////////////////////

static void
finished_CREATED (void) {

  strcpy (header.created, current_cdata);
}


static void
finished_HOST (void) {

  strcpy (header.host, current_cdata);
}


static void
finished_USER (void) {

  strcpy (header.user, current_cdata);
}


static void
finished_METRIC_METRICNAME (void) {

  strcpy (current_metric.metric_name, current_cdata);
 
}


static void
finished_METRICDESCRIPTION (void) {

  strcpy (current_metric.metric_description, current_cdata);
 
}




static void
finished_METRIC_METRICID (void) {

  sscanf (current_cdata, "%d", &current_metric.metric_id);

}





static void
finished_MEASUREMENT_METRICID (void) {

  sscanf (current_cdata, "%d", &current_measurement.metric_id);

}


static void
finished_MEASUREMENT_THREAD (void) {

  sscanf (current_cdata, "%d", &current_measurement.thread_id);

}

static void
finished_MEASUREMENT_OBJECT (void) {

  sscanf (current_cdata, "%d", &current_measurement.object_id);

}


static void
finished_MEASUREMENT_DURATION (void) {

  current_measurement.duration = string_to_ns (current_cdata);

}




static void
finished_MEASUREMENT_BEGIN (void) {

  current_measurement.begin = string_to_ns (current_cdata);

}




static void
finished_MEASUREMENT_END (void) {

  current_measurement.end = string_to_ns (current_cdata);

}




static void
finished_MEASUREMENT_EVCOUNT (void) {

  sscanf (current_cdata, "%d", &current_measurement.events_count);

}




static void
finished_MEASUREMENT_ID (void) {

  sscanf (current_cdata, "%d", &current_measurement.id);

}




static void
finished_METRIC (void) {
  
  strcpy (metric [current_metric.metric_id].metric_name, current_metric.metric_name);
  strcpy (metric [current_metric.metric_id].metric_description, current_metric.metric_description);
  metric [current_metric.metric_id].used = TRUE;

  if (strstr (current_metric.metric_name, "JOB_")            != NULL ||
      strstr (current_metric.metric_name, "RUNNING_GLOBAL")  != NULL ||
      strstr (current_metric.metric_name, "MUTEX_IS_LOCKED") != NULL) {
    metric [current_metric.metric_id].with_metric_summary = FALSE;
  }

  if (strstr (current_metric.metric_name, "MUTEX_IS_LOCKED") != NULL) {
    metric [current_metric.metric_id].show_object = TRUE;
  }

  if (strstr (current_metric.metric_name, "TOTAL_TIME_ELAPSED") != NULL) {
    metric [current_metric.metric_id].with_specific_distributions = FALSE;
  }
}




static void
finished_MEASUREMENT (void) {

  int                  metric_id;
  int                  thread_id;
  int                  object_id;
  measurement_t      * m;
  measurement_list_t * list;
  double               s_x;

  metric_id = current_measurement.metric_id;
  thread_id = current_measurement.thread_id;
  object_id = current_measurement.object_id;

  if (thread_id == NO_THREAD) {
    // no thread is accounted in thread 0
    thread_id = 0;
  }
  if (object_id == NO_THREAD) {
    // no object is accounted in thread 0
    object_id = 0;
  }

  m = (measurement_t *) malloc (sizeof(measurement_t));

  *m = current_measurement;

  list = &(metric [metric_id].measurement_list [object_id][thread_id]);
  if (list->first == NULL ) {
    list->first = m;
  } else {
    list->last->next = m;
  }

  list->last = m;
  m->next = NULL;
    
  list->dist.n      = list->dist.n + 1;
  list->dist.sum_x  = list->dist.sum_x + m->duration;
  s_x = (double) m->duration / 1000000000.0;
  list->dist.sum_x2 = list->dist.sum_x2 + (s_x * s_x);

  if (m->duration < list->dist.min) {
    list->dist.min = m->duration;
  }

  if (m->duration > list->dist.max) {
    list->dist.max = m->duration;
  }

  metric [metric_id].metric_dist.n      = metric [metric_id].metric_dist.n + 1;
  metric [metric_id].metric_dist.sum_x  = metric [metric_id].metric_dist.sum_x  + m->duration;
  metric [metric_id].metric_dist.sum_x2 = metric [metric_id].metric_dist.sum_x2 + (s_x * s_x);
    
  if (m->duration < metric [metric_id].metric_dist.min) {
    metric [metric_id].metric_dist.min = m->duration;
  }
    
  if (m->duration > metric [metric_id].metric_dist.max) {
    metric [metric_id].metric_dist.max = m->duration;
  }
}






//////////////////////////////////////////////////////////////
//  HANDLERS
//////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////
//  startElement
//////////////////////////////////////////////////////////////

static void
startElement(void *userData, const char *name, const char **atts)
{
  int *depthPtr = userData;
  *depthPtr += 1;

  current_cdata_len = 0;
  current_cdata [0] = '\0';

  if (strcmp (name, "METRICSREQUESTED") == 0) {
    start_METRICSREQUESTED ();

  } else if (strcmp (name, "MEASUREMENTS") == 0) {
    start_MEASUREMENTS ();
  }

}




//////////////////////////////////////////////////////////////
//  endElement
//////////////////////////////////////////////////////////////

static void
endElement(void *userData, const char *name)
{
  int *depthPtr = userData;

  *depthPtr -= 1;

  //  for (i = 0; i < *depthPtr; i++)
  //    putchar('\t');
  //  puts(name);


  current_cdata [current_cdata_len] = '\0';
  


  if (strcmp (name, "METRICID") == 0) {
    if (in_MEASUREMENTS == TRUE) {
      finished_MEASUREMENT_METRICID ();
    } else {
      finished_METRIC_METRICID ();
    }
      
  } else if (strcmp (name, "METRICNAME") == 0) {
    if (in_MEASUREMENTS == TRUE) {
      finished_MEASUREMENT_THREAD ();
    } else {
      finished_METRIC_METRICNAME ();
    }

  } else if (strcmp (name, "METRICDESCRIPTION") == 0) {
    finished_METRICDESCRIPTION ();

  } else if (strcmp (name, "THREAD") == 0) {
    finished_MEASUREMENT_THREAD ();

  } else if (strcmp (name, "DURATION") == 0) {
    finished_MEASUREMENT_DURATION ();

  } else if (strcmp (name, "BEGIN") == 0) {
    finished_MEASUREMENT_BEGIN ();

  } else if (strcmp (name, "END") == 0) {
    finished_MEASUREMENT_END ();

  } else if (strcmp (name, "EVCOUNT") == 0) {
    finished_MEASUREMENT_EVCOUNT ();

  } else if (strcmp (name, "ID") == 0) {
    finished_MEASUREMENT_ID ();

  } else if (strcmp (name, "MEASUREMENT") == 0) {
    finished_MEASUREMENT ();

  } else if (strcmp (name, "METRIC") == 0) {
    finished_METRIC ();

  } else if (strcmp (name, "CREATED") == 0) {
    finished_CREATED ();

  } else if (strcmp (name, "HOST") == 0) {
    finished_HOST ();

  } else if (strcmp (name, "USER") == 0) {
    finished_USER ();
  
  } else if (strcmp (name, "OBJECT") == 0) {
    finished_MEASUREMENT_OBJECT ();
  }
  

}




//////////////////////////////////////////////////////////////
//  charData
//////////////////////////////////////////////////////////////

static void
charData(void *userData, const char *cdata, int len)
{
  int i;
  int j = 0;


  i = current_cdata_len;
  while (j < len) {
      current_cdata [i] = cdata [j];
      i++;
      j++;
  }
  current_cdata_len = i;
  
}




//////////////////////////////////////////////////////////////
//  main
//////////////////////////////////////////////////////////////

int
main(int argc, char *argv[])
{
  char buf[BUFSIZ];
  XML_Parser parser = XML_ParserCreate(NULL);
  int done;
  int depth = 0;
  FILE * input;
  int option_input = FALSE;
  int option_dir    = FALSE;
  int i;

  initialize ();

  for (i = 1; i < argc; i++) {

    if (option_input == TRUE) {
      strcpy (inputfilename, argv [i]);
      option_input = FALSE;

    } else if (option_dir == TRUE) {
      strcpy (dir, argv [i]);
      option_dir = FALSE;

    } else if (strcmp (argv [i], "-m") == 0) {
      must_make_measurements = TRUE;

    } else if (strcmp (argv [i], "-d") == 0) {
      option_dir = TRUE;

    } else if (strcmp (argv [i], "-i") == 0) {
      option_input = TRUE;

    } else if (strcmp (argv [i], "--help") == 0) {
      usage (argv [0]);
      
    }  else {
      usage (argv [0]);
    }   
  }
  
  if (strcmp (inputfilename, "") == 0) {
    input = stdin;
  } else {
    input = fopen (inputfilename,"r");
    
    if (input == NULL) {
      fprintf (stderr, "error opening input file %s\n", inputfilename);
      usage (argv [0]);
    }
  }

  mkdir (dir, 0777);
  strcpy (dir_figs, dir);
  strcat (dir_figs, "/figs");
  mkdir (dir_figs, 0777);

  strcpy (reportfilename, dir);
  strcat (reportfilename, "/metricsreport.xml");

  report = fopen (reportfilename, "w+");

  if (report == NULL) {
    fprintf (stderr, "error creating %s\n", reportfilename);
    usage (argv [0]);
  }

  strcpy (progfilename, dir);
  strcat (progfilename, "/prog.plt");

  strcpy (gnuplot_command, "gnuplot ");
  strcat (gnuplot_command, progfilename);


  XML_SetUserData (parser, &depth);
  XML_SetElementHandler (parser, startElement, endElement);
  XML_SetElementHandler (parser, startElement, endElement);

  XML_SetCharacterDataHandler (parser, charData);

  do {
    size_t len = fread(buf, 1, sizeof(buf), input);
    done = len < sizeof(buf);
    if (XML_Parse(parser, buf, len, done) == XML_STATUS_ERROR) {
      fprintf(stderr,
              "%s at line %d\n",
              XML_ErrorString(XML_GetErrorCode(parser)),
              XML_GetCurrentLineNumber(parser));
      return 1;
    }
  } while (!done);



 
  calculate_distributions ();

  make_docbook ();

  XML_ParserFree(parser);
  return 0;
}
